
// symbreg.cc

/* ---------------------------------------------------------------

Symbolic Regression

An example for how to use gpc++ - The Genetic Programming Kernel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Copyright 1996, 1997 Thomas Weinbrenner

For comments, improvements, additions (or even money !?) contact:

Thomas Weinbrenner
Grauensteinstr. 26
35789 Laimbach
Germany
E-mail: thomasw@emk.e-technik.th-darmstadt.de
WWW:    http://www.emk.e-technik.th-darmstadt/~thomasw

--------------------------------------------------------------- */

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>

#include <stdiostream.h>
#include <stdlib.h>
#include <new.h>    // For the new-handler
#include <math.h>   // fabs()
#include <string.h>
#include <getopt.h>

// Include header file of genetic programming system.
#include "gp.h" 
#include "gpconfig.h"
#include "gpd.h"

/* Useful definitions */
#define min(x,y)    (x < y)?x:y
#define max(x,y)    (x < y)?y:x

/* Declare functions to be used later */
char convertToChar(int n);

/* Define the handles for the files so that they can be 
 * used at other places */
ifstream trainin, testin;
int num;

// Define the function to be symbolically regressed and set up
// question and answer array with some points for the function


// The TeX-file
ofstream tout;
int printTexStyle=0;
double penalty;

// Define configuration parameters and the neccessary array to
// read/write the configuration to a file. If you need more variables,
// just add them below and insert an entry in the configArray.
GPVariables cfg;
char *InfoFileName="data";
struct GPConfigVarInformation configArray[]=
{
  {"PopulationSize", DATAINT, &cfg.PopulationSize},
  {"NumberOfGenerations", DATAINT, &cfg.NumberOfGenerations},
  {"CreationType", DATAINT, &cfg.CreationType},
  {"CrossoverProbability", DATADOUBLE, &cfg.CrossoverProbability},
  {"CreationProbability", DATADOUBLE, &cfg.CreationProbability},
  {"MaximumDepthForCreation", DATAINT, &cfg.MaximumDepthForCreation},
  {"MaximumDepthForCrossover", DATAINT, &cfg.MaximumDepthForCrossover},
  {"SelectionType", DATAINT, &cfg.SelectionType},
  {"TournamentSize", DATAINT, &cfg.TournamentSize},
  {"DemeticGrouping", DATAINT, &cfg.DemeticGrouping},
  {"DemeSize", DATAINT, &cfg.DemeSize},
  {"DemeticMigProbability", DATADOUBLE, &cfg.DemeticMigProbability},
  {"SwapMutationProbability", DATADOUBLE, &cfg.SwapMutationProbability},
  {"ShrinkMutationProbability", DATADOUBLE, &cfg.ShrinkMutationProbability},
  {"SteadyState", DATAINT, &cfg.SteadyState},
  {"AddBestToNewPopulation", DATAINT, &cfg.AddBestToNewPopulation},
  {"InfoFileName", DATASTRING, &InfoFileName},
  {"", DATAINT, NULL}
};



// Print out a gene in typical math style. Don't be confused, I don't
// make a difference whether this gene is the main program or an ADF,
// I assume the internal structure is correct.
void MyGene::printMathStyle (ostream& os, int lastPrecedence)
{
  int precedence;

  // Function or terminal?
  if (isFunction ())
    {
      // Determine operator priority
      switch (node->value ())
	{
	case '*': 
	case '%': 
	case 'm':
	case 'x':
	case '=':
	case '<':
	case '>':
	case 's':
	  precedence=1;
	  break;
	case '+': 
	case '-': 
	  precedence=0;
	  break;
	default:
	  GPExitSystem ("MyGene::printMathStyle", 
			"Undefined function value");
	}

      // Do we need brackets?
      if (lastPrecedence>precedence)
	os << "(";

      // Print out the operator and the parameters 
      switch (node->value ())
	{
	case '*': 
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << "*";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  break;
	case '+': 
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << "+";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  break;
	case '-': 
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << "-";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  break;
	case '%': 
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << "%";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  break;
	  /* Our additions */
	case 'm': 
	  os << "min(";
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << ",";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  os << ")";
	  break;
	case 'x': 
	  os << "max(";
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << ",";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  os << ")";
	  break;
	case '=': 
	  os << "(";
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << "==";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  os << ")";
	  break;
	case '<': 
	  os << "(";
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << "<";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  os << ")";
	  break;
	case '>': 
	  os << "(";
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << ">";
	  NthMyChild(1)->printMathStyle (os, precedence);
	  os << ")";
	  break;
	case 's': 
	  os << "sqrt(";
	  NthMyChild(0)->printMathStyle (os, precedence);
	  os << ")";
	  break;
	default: 
	  GPExitSystem ("MyGene::printMathStyle", 
			"Undefined function value");
	}

      // Do we need brackets?
      if (lastPrecedence>precedence)
	os << ")";
    }

  // Print the terminal
  if (isTerminal ())
    os << *node;
}



// Print out a gene in LaTeX-style. Don't be confused, I don't make a
// difference whether this gene is the main program or an ADF, I
// assume the internal structure is correct.
void MyGene::printTeXStyle (ostream& os, int lastPrecedence)
{
  int precedence=0;

  // Function or terminal?
  if (isFunction ())
    {
      // Determine operator priority
      switch (node->value())
	{
	case '*': 
	case '%': 
	case 'm':
	case 'x':
	case '=':
	case '<':
	case '>':
	case 's':
	  precedence=2;
	  break;
	case '+': 
	case '-': 
	  precedence=1;
	  break;
	default:
	  GPExitSystem ("MyGene::printTeXStyle", 
			"Undefined function value");
	}

      // Do we need brackets?
      if (lastPrecedence>precedence)
	os << "\\left(";

      // Print out the operator and the parameters 
      switch (node->value())
	{
	case '*': 
	  NthMyChild(0)->printTeXStyle (os, precedence);
	  os << " ";
	  NthMyChild(1)->printTeXStyle (os, precedence);
	  break;
	case '+': 
	  NthMyChild(0)->printTeXStyle (os, precedence);
	  os << "+";
	  NthMyChild(1)->printTeXStyle (os, precedence);
	  break;
	case '-': 
	  NthMyChild(0)->printTeXStyle (os, precedence);
	  os << "-";
	  NthMyChild(1)->printTeXStyle (os, precedence);
	  break;
	case '%': 
	  // As we use \frac, we start again with precedence 0
	  os << "\\frac{";
	  NthMyChild(0)->printTeXStyle (os, 0);
	  os << "}{";
	  NthMyChild(1)->printTeXStyle (os, 0);
	  os << "}";
	  break;
	case 'm': 
	  os << "min(";
	  NthMyChild(0)->printTeXStyle (os, 0);
	  os << ",";
	  NthMyChild(1)->printTeXStyle (os, 0);
	  os << ")";
	  break;
	case 'x': 
	  os << "max(";
	  NthMyChild(0)->printTeXStyle (os, 0);
	  os << ",";
	  NthMyChild(1)->printTeXStyle (os, 0);
	  os << ")";
	  break;
	case '=': 
	  os << "(";
	  NthMyChild(0)->printTeXStyle (os, 0);
	  os << " == ";
	  NthMyChild(1)->printTeXStyle (os, 0);
	  os << ")";
	  break;
	case '<': 
	  os << "(";
	  NthMyChild(0)->printTeXStyle (os, 0);
	  os << " < ";
	  NthMyChild(1)->printTeXStyle (os, 0);
	  os << ")";
	  break;
	case '>': 
	  os << "(";
	  NthMyChild(0)->printTeXStyle (os, 0);
	  os << " > ";
	  NthMyChild(1)->printTeXStyle (os, 0);
	  os << ")";
	  break;
	case 's': 
	  os << "sqrt(";
	  NthMyChild(0)->printTeXStyle (os, 0);
	  os << ")";
	  break;

	default: 
	  GPExitSystem ("MyGene::printTeXStyle", 
			"Undefined function value");
	}

      // Do we need brackets?
      if (lastPrecedence>precedence)
	os << "\\right)";
    }

  // We can't let the terminal print itself, because we want to modify
  // it a little bit
  if (isTerminal ())
    os << node->value();
}



// Print a Gene.
void MyGene::printOn (ostream& os)
{
  if (printTexStyle)
    printTeXStyle (os);
  else
    printMathStyle (os);
}



// Print a GP. If we want a LaTeX-output, we must provide for the
// equation environment, otherwise we simply call the print function
// of our base class.
void MyGP::printOn (ostream& os)
{
  // If we use LaTeX-style, we provide here for the right equation
  // overhead used for LaTeX. 
  if (printTexStyle)
    {
      tout << "\\begin{eqnarray}" << endl;

      // Print all ADF's, if any
      GPGene* current;
      for (int n=0; n<containerSize(); n++)
	{
	  if (n!=0)
	    os << "\\\\" << endl;
	  os << "f_" << n+1 << " & = & ";
	  if ((current=NthGene (n)))
	    os << *current;
	  else
	    os << " NONE";
	  os << "\\nonumber ";
	}
      tout << endl << "\\end{eqnarray}" << endl << endl;
      
    }
  else
    // Just call the print function of our base class to do the
    // standard job.
    GP::printOn (os);
}



// This function is the divide with closure. Basically if you divide
// anything by zero you get an error so we have to stop this
// process. We check for a very small denominator and return a very
// high value.
inline double divide (double x, double y)
{
  if (fabs (y)<1e-6)
    {
      if (x*y<0.0) 
	return -1e6;
      else
	return 1e6;
    }
  else
    return x/y;
}



// We have the freedom to define this function in any way we like. In
// this case, it takes the parameter x that represents the terminal X,
// and returns the value of the expression. It's recursive of course.
double MyGene::evaluate (tuple &x, MyGP& gp)
{
  double ret;

  if (isFunction ())
    switch (node->value ())
      {
      case '*': 
	ret=NthMyChild(0)->evaluate (x, gp)
	  * NthMyChild(1)->evaluate (x, gp);
	break;
      case '+': 
	ret=NthMyChild(0)->evaluate (x, gp)
	  + NthMyChild(1)->evaluate (x, gp);
	break;
      case '-': 
	ret=NthMyChild(0)->evaluate (x, gp)
	  - NthMyChild(1)->evaluate (x, gp);
	break;
      case '%': 
	// We use the function divide rather than "/" to ensure the
	// closure property
	ret=divide (NthMyChild(0)->evaluate (x, gp), 
		    NthMyChild(1)->evaluate (x, gp));
	break;
	/* Our our additions to the basic functions */
      case 'm': 
	ret=min(NthMyChild(0)->evaluate (x, gp),
	  NthMyChild(1)->evaluate (x, gp));
	break;

      case 'x': 
	ret=max(NthMyChild(0)->evaluate (x, gp),
	  NthMyChild(1)->evaluate (x, gp));
	break;

      case '=': 
	ret=(NthMyChild(0)->evaluate (x, gp) == 
	  NthMyChild(1)->evaluate (x, gp));
	break;

      case '<': 
	ret=(NthMyChild(0)->evaluate (x, gp) < 
	  NthMyChild(1)->evaluate (x, gp));
	break;

      case '>': 
	ret=(NthMyChild(0)->evaluate (x, gp) > 
	  NthMyChild(1)->evaluate (x, gp));
	break;

      case 's': 
	ret=sqrt(NthMyChild(0)->evaluate (x, gp));
	break;
 
      default: 
	GPExitSystem ("MyGene::evaluate", "Undefined function value");
      }
  if (isTerminal ()){
    if (node->value() > 0 && node->value() <= num){
      ret = x.getValue(node->value());
      if (ret == 0){
	/* SOmething bad happened */
	ret = 1;
      }
    }
    else
      GPExitSystem ("MyGene::evaluate", "Undefined terminal value");
  }

  // Restrict the return value (it may become really large, especially
  // for large trees)
  const double maxValue=1e6;
  if (ret>maxValue)
    return maxValue;
  if (ret<-maxValue)
    return -maxValue;

  return ret;
}



// Evaluate the fitness of a GP and save it into the GP class variable
// stdFitness.
void MyGP::evaluate ()
{
  double rawfitness=0.0, diff=0.0;

  // Look at all documents in the training 
  // file to see if we can find the 

  while (!trainin.eof()){
    tuple t;
    if (t.makeTuple(num, trainin) == 0)
      /* finsihed with the records?? */
      break;
    
    double result=NthMyGene(0)->evaluate(t, *this);
    if (result > 0)
      result = 1;
    else
      result = 0;

    if (t.tclass == tuple::ANOMOLY && result == 1)
      rawfitness += penalty;
    else if (t.tclass == tuple::OK && result == 0)
      rawfitness ++;
  }
  
  // Go to the begining of the file again
  trainin.clear();
  trainin.seekg(0);

  stdFitness=sqrt(rawfitness);
  // cout << "std fitness:" << stdFitness << endl;
}



// Create function and terminal set
void createNodeSet (GPAdfNodeSet& adfNs, int n)
{
  // Reserve space for the node sets
  adfNs.reserveSpace (1);

  // Now define the function and terminal set for each ADF and place
  // function/terminal sets into overall ADF container
  GPNodeSet& ns0=*new GPNodeSet (10+n);
  adfNs.put (0, ns0);

  // Define functions/terminals and place them into the appropriate
  // sets. Terminals take two arguments, functions three (the third
  // parameter is the number of arguments the function has)

  ns0.putNode (*new GPNode ('+', "+", 2));
  ns0.putNode (*new GPNode ('-', "-", 2));
  ns0.putNode (*new GPNode ('*', "*", 2));
  ns0.putNode (*new GPNode ('%', "%", 2));
  /* Some new functions */
  ns0.putNode (*new GPNode ('m', "MIN", 2));
  ns0.putNode (*new GPNode ('x', "MAX", 2));
  ns0.putNode (*new GPNode ('=', "=", 2));
  ns0.putNode (*new GPNode ('<', "<", 2));
  ns0.putNode (*new GPNode ('>', ">", 2));
  ns0.putNode (*new GPNode ('s', "SQRT", 1));

  /* Now the terminals, these are the number of items in the tuple*/
  for (int i=1; i<= n; i++){
    char number[3];
    sprintf(number, "%d", i);
    
    ns0.putNode (*new GPNode (i, number));
  }
}



void newHandler ()
{
  cerr << "\nFatal error: Out of memory." << endl;
  exit (1);
}



int main (int argc, char** argv)
{
  int c, len=0;
  char *trainfile=NULL, *testfile=NULL;
  penalty = 1000;
  // We set up a new-handler, because we might need a lot of memory,
  // and we don't know it's there.
  set_new_handler (newHandler);

  /* Handle the command line options */
  while ((c =getopt(argc, argv, "n:f:t:p:")) != -1){
    switch (c) {
    case 'n': {
      /* tuple size */
      num = atoi(optarg);
      break;
    }
    case 'f': {
      len = strlen(optarg);
      trainfile = new char[len + 1];
      if (!trainfile){
        cout << "Out of mem" << endl;
        exit(1);
      }
      strncpy(trainfile, optarg, len);
      trainfile[len] = '\0';
      break;
    }
    case 't': {
      len = strlen(optarg);
      testfile = new char[len + 1];
      if (!testfile){
        cout << "Out of mem" << endl;
        exit(1);
      }
      strncpy(testfile, optarg, len);
      testfile[len] = '\0';
      break;
    }
    case 'p': {
      penalty = atoi(optarg);
      break;
    }
    default:
      cout << "Wrong arguments: -n size -f train -t test -o output" << endl;
      return 1;
    }
  }

  /* Initialize the file pointers */
  if (!trainfile || !testfile){
    cout << "bad training or test filename" << endl;
    return 1;
  }
    
  trainin.open(trainfile);
  if (!trainin.good()){
    cout << "Could not open training file" << endl;
    return (1);
  }
  testin.open(testfile);
  if (!testin.good()){
    cout << "Could not open test file" << endl;
    return (1);
  }

  // Init GP system.
  GPInit (1, -1);

  // Declare the GP Variables, set defaults and read configuration
  // file.  The defaults will be overwritten by the configuration file
  // when read.  If it doesn't exist, the defaults will be written to
  // the file.
  GPConfiguration config (cout, "gpd.ini", configArray);

  // Open the main output file for data and statistics file. First set
  // up names for data file. We use also a TeX-file where the
  // equations are written to in TeX-style. Very nice to look at!
  // Remember we should delete the string from the stream, well just a
  // few bytes
  ostrstream strOutFile, strStatFile, strTeXFile;
  strOutFile  << InfoFileName << ".dat" << ends;
  strStatFile << InfoFileName << ".stc" << ends;
  strTeXFile  << InfoFileName << ".tex" << ends;
  ofstream fout (strOutFile.str());
  ofstream bout (strStatFile.str());
  tout.open (strTeXFile.str(), ios::out);
  tout << endl
       << "\\documentstyle[a4]{article}" << endl
       << "\\begin{document}" << endl;
  
  // Print the configuration to the files just opened
  fout << cfg << endl;
  cout << cfg << endl;
  tout << "\\begin{verbatim}\n" << cfg << "\\end{verbatim}\n" << endl;
  
  // Create the adf function/terminal set and print it out.
  GPAdfNodeSet adfNs;
  createNodeSet (adfNs, num);
  cout << adfNs << endl; 
  fout << adfNs << endl;  

  // Create a population with this configuration
  cout << "Creating initial population ..." << endl;
  MyPopulation* pop=new MyPopulation (cfg, adfNs);
  pop->create ();
  cout << "Ok." << endl;
  pop->createGenerationReport (1, 0, fout, bout);
  pop->evaluateGeneration (1, 0, fout, bout);

  // Print the best of generation to the LaTeX-file.
  printTexStyle=1;
  tout << *pop->NthGP (pop->bestOfPopulation);
  printTexStyle=0;

  // This next for statement is the actual genetic programming system
  // which is in essence just repeated reproduction and crossover loop
  // through all the generations .....
  MyPopulation* newPop=NULL;
  for (int gen=1; gen<=cfg.NumberOfGenerations; gen++)
    {
      // Create a new generation from the old one by applying the
      // genetic operators
      if (!cfg.SteadyState)
	newPop=new MyPopulation (cfg, adfNs);
      pop->generate (*newPop);
      
      // Delete the old generation and make the new the old one
      if (!cfg.SteadyState)
	{
	  delete pop;
	  pop=newPop;
	}

      // Print the best of generation to the LaTeX-file.
      printTexStyle=1;
      tout << "Generation " << gen << ", fitness " 
	   << pop->NthGP (pop->bestOfPopulation)->getFitness()
	   << endl;
      tout << *pop->NthGP (pop->bestOfPopulation);
      printTexStyle=0;

      // Create a report of this generation and how well it is doing
      pop->createGenerationReport (0, gen, fout, bout); 
      pop->evaluateGeneration (0, gen, fout, bout);
    }

  // TeX-file: end of document
  tout << endl
       << "\\end{document}"
       << endl;
  tout.close ();

  cout << "\nResults are in " 
       << InfoFileName << ".dat," 
       << InfoFileName << ".tex,"
       << InfoFileName << ".stc." << endl;

  return 0;
}

/* To convert numbers back to characters...should be inverse of
 * convertToNumber(char) in the tuple class!! */
char convertToChar(int n){
  switch (n) {
  case -3:
    return 'A';
  case -2:
    return 'B';
  case -1:
    return 'C';
    case 1:
      return 'D';
  case 2:
    return 'E';
  case 3:
    return 'F';
  default:
    cout << " Wrong number for a symbol" << endl;
    return 'A';
  }
}

/* evaluate each generation of GPs */
void MyPopulation::evaluateGeneration(int printLegend, int generation,
				      ostream& fout, ostream& bout){

  float false_alarms=0.0, false_positives=0.0, hits=0.0, misses=0.0;
  double fit = 0.0;

  NthMyGP(bestOfPopulation)->scoreOnTest(false_positives, misses, 
					 false_alarms, hits, fit);
      cout << "hits|false_alarms|misses|false_positives|Fitness\n";
      bout << "hits|false_alarms|misses|false_positives|Fitness\n";
  bout << hits << "% " << false_alarms << "% " << misses << "% "
       << false_positives << "% " << fit 
       << endl << endl;
  cout << hits << "% " << false_alarms << "% " << misses << "% "
       << false_positives << "% " << fit
       << endl << endl;

  // Place the best of generation in output files
  fout << "(hits false_alarms misses false_positives fitness" << endl
       << hits << "% " << false_alarms << "% " << misses << "% " 
       << false_positives << "% " << fit << endl << endl;
}

/* score a gp */
void MyGP::scoreOnTest(float& fp, float& m, float &fa, float &h, double &f){
  double rawfitness=0.0;

  double total_records=0, anomolous=0, hit=0, false_alarm=0;
  // Look at all documents in the training 
  // file to see if we can find the 

  while (!testin.eof()){
    tuple t;
    if (t.makeTuple(num, testin) == 0)
      /* finsihed with the records?? */
      break;
    
    /* Keep track of these for percentages */
    total_records ++;
    if (t.tclass == tuple::ANOMOLY) anomolous++;

    double result=NthMyGene(0)->evaluate(t, *this);
    if (result > 0)
      result = 1;
    else
      result = 0;

    if (t.tclass == tuple::ANOMOLY && result == 1) /* miss */
      rawfitness += penalty;
    else if (t.tclass == tuple::OK && result == 0){ /* false_Alarms */
      rawfitness ++;
      false_alarm ++;
    }
    else if (t.tclass == tuple::ANOMOLY) /* hit!! */
      hit ++;
    else /* tclass = OK, false positive/negative*/
      ;
  }
  
  // Go to the begining of the file again
  testin.clear();
  testin.seekg(0);

  f = sqrt(rawfitness);
  fp = (total_records - false_alarm)/total_records * 100; 
  m  = (anomolous - hit)/anomolous * 100;
  fa = false_alarm/total_records * 100;
  h  = hit/anomolous * 100;
  //cout << "total_records" << total_records << endl;
}

